#!python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.


from xpcom import components
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, getProxyForObject
import re

def getProxiedEffectivePrefs(request):
    return getProxyForObject(None,
                             components.interfaces.koIPreferenceSet,
                             request.koDoc.getEffectivePrefs(),
                             PROXY_ALWAYS | PROXY_SYNC)

SEV_ERROR = 2   # No xpcom here :(
SEV_WARNING = 1
SEV_INFO = 0

def createAddResult(results, textlines, severity, lineNo, desc, leadingWS=None):
    """
    lineNo is 1-based
    """
    if lineNo > len(textlines):
        lineNo = len(textlines)
    while lineNo >= 0 and len(textlines[lineNo - 1]) == 0:
        lineNo -= 1
    if lineNo == 0:
        return
    targetLine = textlines[lineNo - 1]
    if leadingWS is not None:
        columnEndOffset = len(leadingWS)
    else:
        columnEndOffset = 0
    columnEnd = len(targetLine) + 1 - columnEndOffset

    result = KoLintResult(description=desc,
                          severity=severity,
                          lineStart=lineNo,
                          lineEnd=lineNo,
                          columnStart=1,
                          columnEnd=columnEnd)
    results.addResult(result)
    
class KoLintResult:
    _com_interfaces_ = [components.interfaces.koILintResult]
# This object is never actually registered and created by contract ID.
# Our language linters create them explicitly.
#    _reg_desc_ = "Komodo Lint Result"
#    _reg_clsid_ = "{21648850-492F-11d4-AC24-0090273E6A60}"
#    _reg_contractid_ = "Komodo.LintResult"

    SEV_INFO    = components.interfaces.koILintResult.SEV_INFO
    SEV_WARNING = components.interfaces.koILintResult.SEV_WARNING
    SEV_ERROR   = components.interfaces.koILintResult.SEV_ERROR

    def __init__(self, description="", severity=None, lineStart=-1, lineEnd=-1, columnStart=-1, columnEnd=-1):
        self.lineStart = lineStart
        self.lineEnd = lineEnd
        self.columnStart = columnStart
        self.columnEnd = columnEnd
        self.description = description
        self.encodedDescription = None
        self.severity = severity

    def __str__(self):
        return "%d:%d (%d-%d) %s" % (
            self.lineStart,
            self.lineEnd,
            self.columnStart,
            self.columnEnd,
        self.description)

    def _encode_string(self, s):
        return re.sub(# match 0x00-0x1f except TAB(0x09), LF(0x0A), and CR(0x0D)
                   '([\x00-\x08\x0b\x0c\x0e-\x1f])',
                   # replace with XML decimal char entity, e.g. '&#7;'
                   lambda m: '\\x%02X'%ord(m.group(1)),
                   s)

    # XXX since this object is not created through xpcom createinstance,
    # using a setter doesn't work.  However, we access it through xpcom
    # from javascript, so we can do our encoding once in the getter
    # to prevent UI crashes.
    def get_description(self):
        if self.encodedDescription is None:
            _gEncodingServices = components.classes['@activestate.com/koEncodingServices;1'].\
                     getService(components.interfaces.koIEncodingServices)
            try:
                unicodebuffer, encoding, bom = _gEncodingServices.\
                                                 getUnicodeEncodedStringUsingOSDefault(self.description)
                self.encodedDescription  = self._encode_string(unicodebuffer)
            except Exception, e:
                self.encodedDescription  = repr(self.description)[1:-1] # remove quotes
        return self.encodedDescription
